# coding: utf-8
"""
    xkbcommon
    ~~~~~~~~~

    CFFI-based libxkbcommon bindings for Python.  See README.rst for details.

    :copyright: Copyright 2016 by Stephen Early
    :license: MIT, see LICENSE for details.
"""
